""" Implementation of binary operators for the different types.
    Currently implemented operators:
        arith: add, sub, mul, truediv, floordiv, mod, pow
        cmp: lt, le, eq, ne, ge, gt
"""
import operator
import numba
from numba.core import types
from numba.core.imputils import lower_builtin
from numba.core.typing.builtins import machine_ints
from numba.core.typing.templates import AbstractTemplate, infer_global
from numba.extending import overload
import bodo
from bodo.hiframes.datetime_date_ext import datetime_date_array_type, datetime_date_type, datetime_timedelta_type
from bodo.hiframes.datetime_timedelta_ext import datetime_datetime_type, datetime_timedelta_array_type, pd_timedelta_type
from bodo.hiframes.pd_dataframe_ext import DataFrameType
from bodo.hiframes.pd_index_ext import DatetimeIndexType, HeterogeneousIndexType, is_index_type
from bodo.hiframes.pd_offsets_ext import date_offset_type, month_begin_type, month_end_type, week_type
from bodo.hiframes.pd_timestamp_ext import pd_timestamp_type
from bodo.hiframes.series_impl import SeriesType
from bodo.libs.binary_arr_ext import binary_array_type, bytes_type
from bodo.libs.bool_arr_ext import boolean_array
from bodo.libs.decimal_arr_ext import Decimal128Type
from bodo.libs.int_arr_ext import IntegerArrayType
from bodo.libs.str_ext import string_type
from bodo.utils.typing import BodoError, is_overload_bool, is_str_arr_type, is_timedelta_type


class SeriesCmpOpTemplate(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        assert len(args) == 2
        lhs, rhs = args
        if cmp_timeseries(lhs, rhs) or (isinstance(lhs, DataFrameType) or
            isinstance(rhs, DataFrameType)) or not (isinstance(lhs,
            SeriesType) or isinstance(rhs, SeriesType)):
            return
        msrp__dyk = lhs.data if isinstance(lhs, SeriesType) else lhs
        ojdvt__vws = rhs.data if isinstance(rhs, SeriesType) else rhs
        if msrp__dyk in (bodo.pd_timestamp_type, bodo.pd_timedelta_type
            ) and ojdvt__vws.dtype in (bodo.datetime64ns, bodo.timedelta64ns):
            msrp__dyk = ojdvt__vws.dtype
        elif ojdvt__vws in (bodo.pd_timestamp_type, bodo.pd_timedelta_type
            ) and msrp__dyk.dtype in (bodo.datetime64ns, bodo.timedelta64ns):
            ojdvt__vws = msrp__dyk.dtype
        vafe__tqly = msrp__dyk, ojdvt__vws
        qamfn__qoin = (
            f'{lhs} {numba.core.utils.OPERATORS_TO_BUILTINS[self.key]} {rhs} not supported'
            )
        try:
            ycx__klbak = self.context.resolve_function_type(self.key,
                vafe__tqly, {}).return_type
        except Exception as itp__sczwe:
            raise BodoError(qamfn__qoin)
        if is_overload_bool(ycx__klbak):
            raise BodoError(qamfn__qoin)
        gcnlk__qmww = lhs.index if isinstance(lhs, SeriesType) else rhs.index
        jap__udqn = lhs.name_typ if isinstance(lhs, SeriesType
            ) else rhs.name_typ
        auf__ado = types.bool_
        qpee__iagp = SeriesType(auf__ado, ycx__klbak, gcnlk__qmww, jap__udqn)
        return qpee__iagp(*args)


def series_cmp_op_lower(op):

    def lower_impl(context, builder, sig, args):
        dtkpa__ign = bodo.hiframes.series_impl.create_binary_op_overload(op)(*
            sig.args)
        if dtkpa__ign is None:
            dtkpa__ign = create_overload_cmp_operator(op)(*sig.args)
        return context.compile_internal(builder, dtkpa__ign, sig, args)
    return lower_impl


class SeriesAndOrTyper(AbstractTemplate):

    def generic(self, args, kws):
        assert len(args) == 2
        assert not kws
        lhs, rhs = args
        if not (isinstance(lhs, SeriesType) or isinstance(rhs, SeriesType)):
            return
        msrp__dyk = lhs.data if isinstance(lhs, SeriesType) else lhs
        ojdvt__vws = rhs.data if isinstance(rhs, SeriesType) else rhs
        vafe__tqly = msrp__dyk, ojdvt__vws
        qamfn__qoin = (
            f'{lhs} {numba.core.utils.OPERATORS_TO_BUILTINS[self.key]} {rhs} not supported'
            )
        try:
            ycx__klbak = self.context.resolve_function_type(self.key,
                vafe__tqly, {}).return_type
        except Exception as flbbe__qurne:
            raise BodoError(qamfn__qoin)
        gcnlk__qmww = lhs.index if isinstance(lhs, SeriesType) else rhs.index
        jap__udqn = lhs.name_typ if isinstance(lhs, SeriesType
            ) else rhs.name_typ
        auf__ado = ycx__klbak.dtype
        qpee__iagp = SeriesType(auf__ado, ycx__klbak, gcnlk__qmww, jap__udqn)
        return qpee__iagp(*args)


def lower_series_and_or(op):

    def lower_and_or_impl(context, builder, sig, args):
        dtkpa__ign = bodo.hiframes.series_impl.create_binary_op_overload(op)(*
            sig.args)
        if dtkpa__ign is None:
            lhs, rhs = sig.args
            if isinstance(lhs, DataFrameType) or isinstance(rhs, DataFrameType
                ):
                dtkpa__ign = (bodo.hiframes.dataframe_impl.
                    create_binary_op_overload(op)(*sig.args))
        return context.compile_internal(builder, dtkpa__ign, sig, args)
    return lower_and_or_impl


def overload_add_operator_scalars(lhs, rhs):
    if lhs == week_type or rhs == week_type:
        return (bodo.hiframes.pd_offsets_ext.
            overload_add_operator_week_offset_type(lhs, rhs))
    if lhs == month_begin_type or rhs == month_begin_type:
        return (bodo.hiframes.pd_offsets_ext.
            overload_add_operator_month_begin_offset_type(lhs, rhs))
    if lhs == month_end_type or rhs == month_end_type:
        return (bodo.hiframes.pd_offsets_ext.
            overload_add_operator_month_end_offset_type(lhs, rhs))
    if lhs == date_offset_type or rhs == date_offset_type:
        return (bodo.hiframes.pd_offsets_ext.
            overload_add_operator_date_offset_type(lhs, rhs))
    if add_timestamp(lhs, rhs):
        return bodo.hiframes.pd_timestamp_ext.overload_add_operator_timestamp(
            lhs, rhs)
    if add_dt_td_and_dt_date(lhs, rhs):
        return (bodo.hiframes.datetime_date_ext.
            overload_add_operator_datetime_date(lhs, rhs))
    if add_datetime_and_timedeltas(lhs, rhs):
        return (bodo.hiframes.datetime_timedelta_ext.
            overload_add_operator_datetime_timedelta(lhs, rhs))
    raise_error_if_not_numba_supported(operator.add, lhs, rhs)


def overload_sub_operator_scalars(lhs, rhs):
    if sub_offset_to_datetime_or_timestamp(lhs, rhs):
        return bodo.hiframes.pd_offsets_ext.overload_sub_operator_offsets(lhs,
            rhs)
    if lhs == pd_timestamp_type and rhs in [pd_timestamp_type,
        datetime_timedelta_type, pd_timedelta_type]:
        return bodo.hiframes.pd_timestamp_ext.overload_sub_operator_timestamp(
            lhs, rhs)
    if sub_dt_or_td(lhs, rhs):
        return (bodo.hiframes.datetime_date_ext.
            overload_sub_operator_datetime_date(lhs, rhs))
    if sub_datetime_and_timedeltas(lhs, rhs):
        return (bodo.hiframes.datetime_timedelta_ext.
            overload_sub_operator_datetime_timedelta(lhs, rhs))
    if lhs == datetime_datetime_type and rhs == datetime_datetime_type:
        return (bodo.hiframes.datetime_datetime_ext.
            overload_sub_operator_datetime_datetime(lhs, rhs))
    raise_error_if_not_numba_supported(operator.sub, lhs, rhs)


def create_overload_arith_op(op):

    def overload_arith_operator(lhs, rhs):
        bodo.hiframes.pd_timestamp_ext.check_tz_aware_unsupported(lhs,
            f'{op} operator')
        bodo.hiframes.pd_timestamp_ext.check_tz_aware_unsupported(rhs,
            f'{op} operator')
        if isinstance(lhs, DataFrameType) or isinstance(rhs, DataFrameType):
            return bodo.hiframes.dataframe_impl.create_binary_op_overload(op)(
                lhs, rhs)
        if time_series_operation(lhs, rhs) and op in [operator.add,
            operator.sub]:
            return bodo.hiframes.series_dt_impl.create_bin_op_overload(op)(lhs,
                rhs)
        if isinstance(lhs, SeriesType) or isinstance(rhs, SeriesType):
            return bodo.hiframes.series_impl.create_binary_op_overload(op)(lhs,
                rhs)
        if sub_dt_index_and_timestamp(lhs, rhs) and op == operator.sub:
            return (bodo.hiframes.pd_index_ext.
                overload_sub_operator_datetime_index(lhs, rhs))
        if operand_is_index(lhs) or operand_is_index(rhs):
            return bodo.hiframes.pd_index_ext.create_binary_op_overload(op)(lhs
                , rhs)
        if args_td_and_int_array(lhs, rhs):
            return bodo.libs.int_arr_ext.get_int_array_op_pd_td(op)(lhs, rhs)
        if isinstance(lhs, IntegerArrayType) or isinstance(rhs,
            IntegerArrayType):
            return bodo.libs.int_arr_ext.create_op_overload(op, 2)(lhs, rhs)
        if lhs == boolean_array or rhs == boolean_array:
            return bodo.libs.bool_arr_ext.create_op_overload(op, 2)(lhs, rhs)
        if op == operator.add and (is_str_arr_type(lhs) or types.unliteral(
            lhs) == string_type):
            return bodo.libs.str_arr_ext.overload_add_operator_string_array(lhs
                , rhs)
        if op == operator.add:
            return overload_add_operator_scalars(lhs, rhs)
        if op == operator.sub:
            return overload_sub_operator_scalars(lhs, rhs)
        if op == operator.mul:
            if mul_timedelta_and_int(lhs, rhs):
                return (bodo.hiframes.datetime_timedelta_ext.
                    overload_mul_operator_timedelta(lhs, rhs))
            if mul_string_arr_and_int(lhs, rhs):
                return bodo.libs.str_arr_ext.overload_mul_operator_str_arr(lhs,
                    rhs)
            if mul_date_offset_and_int(lhs, rhs):
                return (bodo.hiframes.pd_offsets_ext.
                    overload_mul_date_offset_types(lhs, rhs))
            raise_error_if_not_numba_supported(op, lhs, rhs)
        if op in [operator.truediv, operator.floordiv]:
            if div_timedelta_and_int(lhs, rhs):
                if op == operator.truediv:
                    return (bodo.hiframes.datetime_timedelta_ext.
                        overload_truediv_operator_pd_timedelta(lhs, rhs))
                else:
                    return (bodo.hiframes.datetime_timedelta_ext.
                        overload_floordiv_operator_pd_timedelta(lhs, rhs))
            if div_datetime_timedelta(lhs, rhs):
                if op == operator.truediv:
                    return (bodo.hiframes.datetime_timedelta_ext.
                        overload_truediv_operator_dt_timedelta(lhs, rhs))
                else:
                    return (bodo.hiframes.datetime_timedelta_ext.
                        overload_floordiv_operator_dt_timedelta(lhs, rhs))
            raise_error_if_not_numba_supported(op, lhs, rhs)
        if op == operator.mod:
            if mod_timedeltas(lhs, rhs):
                return (bodo.hiframes.datetime_timedelta_ext.
                    overload_mod_operator_timedeltas(lhs, rhs))
            raise_error_if_not_numba_supported(op, lhs, rhs)
        if op == operator.pow:
            raise_error_if_not_numba_supported(op, lhs, rhs)
        raise BodoError(
            f'{op} operator not supported for data types {lhs} and {rhs}.')
    return overload_arith_operator


def create_overload_cmp_operator(op):

    def overload_cmp_operator(lhs, rhs):
        if isinstance(lhs, DataFrameType) or isinstance(rhs, DataFrameType):
            bodo.hiframes.pd_timestamp_ext.check_tz_aware_unsupported(lhs,
                f'{op} operator')
            bodo.hiframes.pd_timestamp_ext.check_tz_aware_unsupported(rhs,
                f'{op} operator')
            return bodo.hiframes.dataframe_impl.create_binary_op_overload(op)(
                lhs, rhs)
        if cmp_timeseries(lhs, rhs):
            return bodo.hiframes.series_dt_impl.create_cmp_op_overload(op)(lhs,
                rhs)
        if isinstance(lhs, SeriesType) or isinstance(rhs, SeriesType):
            return
        bodo.hiframes.pd_timestamp_ext.check_tz_aware_unsupported(lhs,
            f'{op} operator')
        bodo.hiframes.pd_timestamp_ext.check_tz_aware_unsupported(rhs,
            f'{op} operator')
        if lhs == datetime_date_array_type or rhs == datetime_date_array_type:
            return bodo.hiframes.datetime_date_ext.create_cmp_op_overload_arr(
                op)(lhs, rhs)
        if (lhs == datetime_timedelta_array_type or rhs ==
            datetime_timedelta_array_type):
            dtkpa__ign = (bodo.hiframes.datetime_timedelta_ext.
                create_cmp_op_overload(op))
            return dtkpa__ign(lhs, rhs)
        if is_str_arr_type(lhs) or is_str_arr_type(rhs):
            return bodo.libs.str_arr_ext.create_binary_op_overload(op)(lhs, rhs
                )
        if isinstance(lhs, Decimal128Type) and isinstance(rhs, Decimal128Type):
            return bodo.libs.decimal_arr_ext.decimal_create_cmp_op_overload(op
                )(lhs, rhs)
        if lhs == boolean_array or rhs == boolean_array:
            return bodo.libs.bool_arr_ext.create_op_overload(op, 2)(lhs, rhs)
        if isinstance(lhs, IntegerArrayType) or isinstance(rhs,
            IntegerArrayType):
            return bodo.libs.int_arr_ext.create_op_overload(op, 2)(lhs, rhs)
        if binary_array_cmp(lhs, rhs):
            return bodo.libs.binary_arr_ext.create_binary_cmp_op_overload(op)(
                lhs, rhs)
        if cmp_dt_index_to_string(lhs, rhs):
            return bodo.hiframes.pd_index_ext.overload_binop_dti_str(op)(lhs,
                rhs)
        if operand_is_index(lhs) or operand_is_index(rhs):
            return bodo.hiframes.pd_index_ext.create_binary_op_overload(op)(lhs
                , rhs)
        if lhs == datetime_date_type and rhs == datetime_date_type:
            return bodo.hiframes.datetime_date_ext.create_cmp_op_overload(op)(
                lhs, rhs)
        if can_cmp_date_datetime(lhs, rhs, op):
            return (bodo.hiframes.datetime_date_ext.
                create_datetime_date_cmp_op_overload(op)(lhs, rhs))
        if lhs == datetime_datetime_type and rhs == datetime_datetime_type:
            return bodo.hiframes.datetime_datetime_ext.create_cmp_op_overload(
                op)(lhs, rhs)
        if lhs == datetime_timedelta_type and rhs == datetime_timedelta_type:
            return bodo.hiframes.datetime_timedelta_ext.create_cmp_op_overload(
                op)(lhs, rhs)
        if cmp_timedeltas(lhs, rhs):
            dtkpa__ign = (bodo.hiframes.datetime_timedelta_ext.
                pd_create_cmp_op_overload(op))
            return dtkpa__ign(lhs, rhs)
        if cmp_timestamp_or_date(lhs, rhs):
            return (bodo.hiframes.pd_timestamp_ext.
                create_timestamp_cmp_op_overload(op)(lhs, rhs))
        if cmp_op_supported_by_numba(lhs, rhs):
            return
        raise BodoError(
            f'{op} operator not supported for data types {lhs} and {rhs}.')
    return overload_cmp_operator


def add_dt_td_and_dt_date(lhs, rhs):
    dwue__cfs = lhs == datetime_timedelta_type and rhs == datetime_date_type
    lbzi__gmue = rhs == datetime_timedelta_type and lhs == datetime_date_type
    return dwue__cfs or lbzi__gmue


def add_timestamp(lhs, rhs):
    kryn__ruy = lhs == pd_timestamp_type and is_timedelta_type(rhs)
    inb__kixub = is_timedelta_type(lhs) and rhs == pd_timestamp_type
    return kryn__ruy or inb__kixub


def add_datetime_and_timedeltas(lhs, rhs):
    rud__kjxbs = [datetime_timedelta_type, pd_timedelta_type]
    vqq__igrci = [datetime_timedelta_type, pd_timedelta_type,
        datetime_datetime_type]
    ehdf__udvx = lhs in rud__kjxbs and rhs in rud__kjxbs
    det__qbklh = (lhs == datetime_datetime_type and rhs in rud__kjxbs or 
        rhs == datetime_datetime_type and lhs in rud__kjxbs)
    return ehdf__udvx or det__qbklh


def mul_string_arr_and_int(lhs, rhs):
    ojdvt__vws = isinstance(lhs, types.Integer) and is_str_arr_type(rhs)
    msrp__dyk = is_str_arr_type(lhs) and isinstance(rhs, types.Integer)
    return ojdvt__vws or msrp__dyk


def mul_timedelta_and_int(lhs, rhs):
    dwue__cfs = lhs in [pd_timedelta_type, datetime_timedelta_type
        ] and isinstance(rhs, types.Integer)
    lbzi__gmue = rhs in [pd_timedelta_type, datetime_timedelta_type
        ] and isinstance(lhs, types.Integer)
    return dwue__cfs or lbzi__gmue


def mul_date_offset_and_int(lhs, rhs):
    blq__ybjw = lhs in [week_type, month_end_type, month_begin_type,
        date_offset_type] and isinstance(rhs, types.Integer)
    mhzs__gvdk = rhs in [week_type, month_end_type, month_begin_type,
        date_offset_type] and isinstance(lhs, types.Integer)
    return blq__ybjw or mhzs__gvdk


def sub_offset_to_datetime_or_timestamp(lhs, rhs):
    enxw__kxm = [datetime_datetime_type, pd_timestamp_type, datetime_date_type]
    cxrn__gouun = [date_offset_type, month_begin_type, month_end_type,
        week_type]
    return rhs in cxrn__gouun and lhs in enxw__kxm


def sub_dt_index_and_timestamp(lhs, rhs):
    ugs__ztf = isinstance(lhs, DatetimeIndexType) and rhs == pd_timestamp_type
    wmbh__fhk = isinstance(rhs, DatetimeIndexType) and lhs == pd_timestamp_type
    return ugs__ztf or wmbh__fhk


def sub_dt_or_td(lhs, rhs):
    pgwb__pry = lhs == datetime_date_type and rhs == datetime_timedelta_type
    ect__jynp = lhs == datetime_date_type and rhs == datetime_date_type
    rvsn__tija = (lhs == datetime_date_array_type and rhs ==
        datetime_timedelta_type)
    return pgwb__pry or ect__jynp or rvsn__tija


def sub_datetime_and_timedeltas(lhs, rhs):
    uwa__ign = (is_timedelta_type(lhs) or lhs == datetime_datetime_type
        ) and is_timedelta_type(rhs)
    zcpt__pqsve = (lhs == datetime_timedelta_array_type and rhs ==
        datetime_timedelta_type)
    return uwa__ign or zcpt__pqsve


def div_timedelta_and_int(lhs, rhs):
    ehdf__udvx = lhs == pd_timedelta_type and rhs == pd_timedelta_type
    czjv__xwnjo = lhs == pd_timedelta_type and isinstance(rhs, types.Integer)
    return ehdf__udvx or czjv__xwnjo


def div_datetime_timedelta(lhs, rhs):
    ehdf__udvx = (lhs == datetime_timedelta_type and rhs ==
        datetime_timedelta_type)
    czjv__xwnjo = lhs == datetime_timedelta_type and rhs == types.int64
    return ehdf__udvx or czjv__xwnjo


def mod_timedeltas(lhs, rhs):
    oli__olouj = lhs == pd_timedelta_type and rhs == pd_timedelta_type
    nae__nnhq = (lhs == datetime_timedelta_type and rhs ==
        datetime_timedelta_type)
    return oli__olouj or nae__nnhq


def cmp_dt_index_to_string(lhs, rhs):
    ugs__ztf = isinstance(lhs, DatetimeIndexType) and types.unliteral(rhs
        ) == string_type
    wmbh__fhk = isinstance(rhs, DatetimeIndexType) and types.unliteral(lhs
        ) == string_type
    return ugs__ztf or wmbh__fhk


def cmp_timestamp_or_date(lhs, rhs):
    zbqg__kacdh = (lhs == pd_timestamp_type and rhs == bodo.hiframes.
        datetime_date_ext.datetime_date_type)
    mpm__qnlh = (lhs == bodo.hiframes.datetime_date_ext.datetime_date_type and
        rhs == pd_timestamp_type)
    vqw__pri = lhs == pd_timestamp_type and rhs == pd_timestamp_type
    snkd__ndw = lhs == pd_timestamp_type and rhs == bodo.datetime64ns
    yykpx__kpfj = rhs == pd_timestamp_type and lhs == bodo.datetime64ns
    return zbqg__kacdh or mpm__qnlh or vqw__pri or snkd__ndw or yykpx__kpfj


def cmp_timeseries(lhs, rhs):
    sjvhp__vyx = bodo.hiframes.pd_series_ext.is_dt64_series_typ(rhs) and (bodo
        .utils.typing.is_overload_constant_str(lhs) or lhs == bodo.libs.
        str_ext.string_type or lhs == bodo.hiframes.pd_timestamp_ext.
        pd_timestamp_type)
    sth__efzw = bodo.hiframes.pd_series_ext.is_dt64_series_typ(lhs) and (bodo
        .utils.typing.is_overload_constant_str(rhs) or rhs == bodo.libs.
        str_ext.string_type or rhs == bodo.hiframes.pd_timestamp_ext.
        pd_timestamp_type)
    drwp__yzg = sjvhp__vyx or sth__efzw
    clbmp__igl = bodo.hiframes.pd_series_ext.is_timedelta64_series_typ(rhs
        ) and lhs == bodo.hiframes.datetime_timedelta_ext.datetime_timedelta_type
    gmj__mxtzk = bodo.hiframes.pd_series_ext.is_timedelta64_series_typ(lhs
        ) and rhs == bodo.hiframes.datetime_timedelta_ext.datetime_timedelta_type
    rdg__ogb = clbmp__igl or gmj__mxtzk
    return drwp__yzg or rdg__ogb


def cmp_timedeltas(lhs, rhs):
    ehdf__udvx = [pd_timedelta_type, bodo.timedelta64ns]
    return lhs in ehdf__udvx and rhs in ehdf__udvx


def operand_is_index(operand):
    return is_index_type(operand) or isinstance(operand, HeterogeneousIndexType
        )


def helper_time_series_checks(operand):
    ezrzb__tpijl = bodo.hiframes.pd_series_ext.is_dt64_series_typ(operand
        ) or bodo.hiframes.pd_series_ext.is_timedelta64_series_typ(operand
        ) or operand in [datetime_timedelta_type, datetime_datetime_type,
        pd_timestamp_type]
    return ezrzb__tpijl


def binary_array_cmp(lhs, rhs):
    return lhs == binary_array_type and rhs in [bytes_type, binary_array_type
        ] or lhs in [bytes_type, binary_array_type
        ] and rhs == binary_array_type


def can_cmp_date_datetime(lhs, rhs, op):
    return op in (operator.eq, operator.ne) and (lhs == datetime_date_type and
        rhs == datetime_datetime_type or lhs == datetime_datetime_type and 
        rhs == datetime_date_type)


def time_series_operation(lhs, rhs):
    uqi__undmf = bodo.hiframes.pd_series_ext.is_timedelta64_series_typ(lhs
        ) and rhs == datetime_timedelta_type
    zwqj__pfwn = bodo.hiframes.pd_series_ext.is_timedelta64_series_typ(rhs
        ) and lhs == datetime_timedelta_type
    nuuu__wdl = bodo.hiframes.pd_series_ext.is_dt64_series_typ(lhs
        ) and helper_time_series_checks(rhs)
    uzn__zwkzz = bodo.hiframes.pd_series_ext.is_dt64_series_typ(rhs
        ) and helper_time_series_checks(lhs)
    return uqi__undmf or zwqj__pfwn or nuuu__wdl or uzn__zwkzz


def args_td_and_int_array(lhs, rhs):
    wgriq__vvo = (isinstance(lhs, IntegerArrayType) or isinstance(lhs,
        types.Array) and isinstance(lhs.dtype, types.Integer)) or (isinstance
        (rhs, IntegerArrayType) or isinstance(rhs, types.Array) and
        isinstance(rhs.dtype, types.Integer))
    magi__yluf = lhs in [pd_timedelta_type] or rhs in [pd_timedelta_type]
    return wgriq__vvo and magi__yluf


def arith_op_supported_by_numba(op, lhs, rhs):
    if op == operator.mul:
        lbzi__gmue = isinstance(lhs, (types.Integer, types.Float)
            ) and isinstance(rhs, types.NPTimedelta)
        dwue__cfs = isinstance(rhs, (types.Integer, types.Float)
            ) and isinstance(lhs, types.NPTimedelta)
        qpukr__nyjz = lbzi__gmue or dwue__cfs
        wgfi__hxwc = isinstance(rhs, types.UnicodeType) and isinstance(lhs,
            types.Integer)
        cchwv__cie = isinstance(lhs, types.UnicodeType) and isinstance(rhs,
            types.Integer)
        hvj__rghdd = wgfi__hxwc or cchwv__cie
        yioc__jgk = isinstance(lhs, types.Integer) and isinstance(rhs,
            types.Integer)
        gfm__wxein = isinstance(lhs, types.Float) and isinstance(rhs, types
            .Float)
        zge__zny = isinstance(lhs, types.Complex) and isinstance(rhs, types
            .Complex)
        bjv__zrym = yioc__jgk or gfm__wxein or zge__zny
        conx__jbzi = isinstance(lhs, types.List) and isinstance(rhs, types.
            Integer) or isinstance(lhs, types.Integer) and isinstance(rhs,
            types.List)
        tys = types.UnicodeCharSeq, types.CharSeq, types.Bytes
        kjyi__dtcs = isinstance(lhs, tys) or isinstance(rhs, tys)
        akblu__gnfjs = isinstance(lhs, types.Array) or isinstance(rhs,
            types.Array)
        return (qpukr__nyjz or hvj__rghdd or bjv__zrym or conx__jbzi or
            kjyi__dtcs or akblu__gnfjs)
    if op == operator.pow:
        msodm__nbkfk = isinstance(lhs, types.Integer) and isinstance(rhs, (
            types.IntegerLiteral, types.Integer))
        mrygt__vrp = isinstance(lhs, types.Float) and isinstance(rhs, (
            types.IntegerLiteral, types.Float, types.Integer) or rhs in
            types.unsigned_domain or rhs in types.signed_domain)
        zge__zny = isinstance(lhs, types.Complex) and isinstance(rhs, types
            .Complex)
        akblu__gnfjs = isinstance(lhs, types.Array) or isinstance(rhs,
            types.Array)
        return msodm__nbkfk or mrygt__vrp or zge__zny or akblu__gnfjs
    if op == operator.floordiv:
        gfm__wxein = lhs in types.real_domain and rhs in types.real_domain
        yioc__jgk = isinstance(lhs, types.Integer) and isinstance(rhs,
            types.Integer)
        lmkza__kdi = isinstance(lhs, types.Float) and isinstance(rhs, types
            .Float)
        ehdf__udvx = isinstance(lhs, types.NPTimedelta) and isinstance(rhs,
            (types.Integer, types.Float, types.NPTimedelta))
        akblu__gnfjs = isinstance(lhs, types.Array) or isinstance(rhs,
            types.Array)
        return (gfm__wxein or yioc__jgk or lmkza__kdi or ehdf__udvx or
            akblu__gnfjs)
    if op == operator.truediv:
        talc__uzxbi = lhs in machine_ints and rhs in machine_ints
        gfm__wxein = lhs in types.real_domain and rhs in types.real_domain
        zge__zny = lhs in types.complex_domain and rhs in types.complex_domain
        yioc__jgk = isinstance(lhs, types.Integer) and isinstance(rhs,
            types.Integer)
        lmkza__kdi = isinstance(lhs, types.Float) and isinstance(rhs, types
            .Float)
        ulkbb__fqe = isinstance(lhs, types.Complex) and isinstance(rhs,
            types.Complex)
        ehdf__udvx = isinstance(lhs, types.NPTimedelta) and isinstance(rhs,
            (types.Integer, types.Float, types.NPTimedelta))
        akblu__gnfjs = isinstance(lhs, types.Array) or isinstance(rhs,
            types.Array)
        return (talc__uzxbi or gfm__wxein or zge__zny or yioc__jgk or
            lmkza__kdi or ulkbb__fqe or ehdf__udvx or akblu__gnfjs)
    if op == operator.mod:
        talc__uzxbi = lhs in machine_ints and rhs in machine_ints
        gfm__wxein = lhs in types.real_domain and rhs in types.real_domain
        yioc__jgk = isinstance(lhs, types.Integer) and isinstance(rhs,
            types.Integer)
        lmkza__kdi = isinstance(lhs, types.Float) and isinstance(rhs, types
            .Float)
        akblu__gnfjs = isinstance(lhs, types.Array) or isinstance(rhs,
            types.Array)
        return (talc__uzxbi or gfm__wxein or yioc__jgk or lmkza__kdi or
            akblu__gnfjs)
    if op == operator.add or op == operator.sub:
        qpukr__nyjz = isinstance(lhs, types.NPTimedelta) and isinstance(rhs,
            types.NPTimedelta)
        sahez__lyj = isinstance(lhs, types.NPDatetime) and isinstance(rhs,
            types.NPDatetime)
        jwg__knjqn = isinstance(lhs, types.NPDatetime) and isinstance(rhs,
            types.NPTimedelta)
        lbkoh__ubq = isinstance(lhs, types.Set) and isinstance(rhs, types.Set)
        yioc__jgk = isinstance(lhs, types.Integer) and isinstance(rhs,
            types.Integer)
        gfm__wxein = isinstance(lhs, types.Float) and isinstance(rhs, types
            .Float)
        zge__zny = isinstance(lhs, types.Complex) and isinstance(rhs, types
            .Complex)
        bjv__zrym = yioc__jgk or gfm__wxein or zge__zny
        akblu__gnfjs = isinstance(lhs, types.Array) or isinstance(rhs,
            types.Array)
        qrx__hlng = isinstance(lhs, types.BaseTuple) and isinstance(rhs,
            types.BaseTuple)
        conx__jbzi = isinstance(lhs, types.List) and isinstance(rhs, types.List
            )
        fxdkn__uhlkp = isinstance(lhs, types.UnicodeCharSeq) and isinstance(rhs
            , types.UnicodeType)
        gvgug__siww = isinstance(rhs, types.UnicodeCharSeq) and isinstance(lhs,
            types.UnicodeType)
        munhu__vkd = isinstance(lhs, types.UnicodeCharSeq) and isinstance(rhs,
            types.UnicodeCharSeq)
        vcn__slg = isinstance(lhs, (types.CharSeq, types.Bytes)
            ) and isinstance(rhs, (types.CharSeq, types.Bytes))
        qnb__qri = fxdkn__uhlkp or gvgug__siww or munhu__vkd or vcn__slg
        hvj__rghdd = isinstance(lhs, types.UnicodeType) and isinstance(rhs,
            types.UnicodeType)
        ioro__lxgj = isinstance(lhs, types.UnicodeType) and isinstance(rhs,
            types.UnicodeCharSeq)
        nhvky__llp = hvj__rghdd or ioro__lxgj
        xoow__pzq = lhs == types.NPTimedelta and rhs == types.NPDatetime
        vgop__ucz = (qrx__hlng or conx__jbzi or qnb__qri or nhvky__llp or
            xoow__pzq)
        taqz__rro = op == operator.add and vgop__ucz
        return (qpukr__nyjz or sahez__lyj or jwg__knjqn or lbkoh__ubq or
            bjv__zrym or akblu__gnfjs or taqz__rro)


def cmp_op_supported_by_numba(lhs, rhs):
    akblu__gnfjs = isinstance(lhs, types.Array) or isinstance(rhs, types.Array)
    conx__jbzi = isinstance(lhs, types.ListType) and isinstance(rhs, types.
        ListType)
    qpukr__nyjz = isinstance(lhs, types.NPTimedelta) and isinstance(rhs,
        types.NPTimedelta)
    pebbs__kbv = isinstance(lhs, types.NPDatetime) and isinstance(rhs,
        types.NPDatetime)
    unicode_types = (types.UnicodeType, types.StringLiteral, types.CharSeq,
        types.Bytes, types.UnicodeCharSeq)
    hvj__rghdd = isinstance(lhs, unicode_types) and isinstance(rhs,
        unicode_types)
    qrx__hlng = isinstance(lhs, types.BaseTuple) and isinstance(rhs, types.
        BaseTuple)
    lbkoh__ubq = isinstance(lhs, types.Set) and isinstance(rhs, types.Set)
    bjv__zrym = isinstance(lhs, types.Number) and isinstance(rhs, types.Number)
    bxit__reg = isinstance(lhs, types.Boolean) and isinstance(rhs, types.
        Boolean)
    otuud__qwi = isinstance(lhs, types.NoneType) or isinstance(rhs, types.
        NoneType)
    qlc__dcwmh = isinstance(lhs, types.DictType) and isinstance(rhs, types.
        DictType)
    euuq__zqqs = isinstance(lhs, types.EnumMember) and isinstance(rhs,
        types.EnumMember)
    zuvxe__txvhb = isinstance(lhs, types.Literal) and isinstance(rhs, types
        .Literal)
    return (conx__jbzi or qpukr__nyjz or pebbs__kbv or hvj__rghdd or
        qrx__hlng or lbkoh__ubq or bjv__zrym or bxit__reg or otuud__qwi or
        qlc__dcwmh or akblu__gnfjs or euuq__zqqs or zuvxe__txvhb)


def raise_error_if_not_numba_supported(op, lhs, rhs):
    if arith_op_supported_by_numba(op, lhs, rhs):
        return
    raise BodoError(
        f'{op} operator not supported for data types {lhs} and {rhs}.')


def _install_series_and_or():
    for op in (operator.or_, operator.and_):
        infer_global(op)(SeriesAndOrTyper)
        lower_impl = lower_series_and_or(op)
        lower_builtin(op, SeriesType, SeriesType)(lower_impl)
        lower_builtin(op, SeriesType, types.Any)(lower_impl)
        lower_builtin(op, types.Any, SeriesType)(lower_impl)


_install_series_and_or()


def _install_cmp_ops():
    for op in (operator.lt, operator.eq, operator.ne, operator.ge, operator
        .gt, operator.le):
        infer_global(op)(SeriesCmpOpTemplate)
        lower_impl = series_cmp_op_lower(op)
        lower_builtin(op, SeriesType, SeriesType)(lower_impl)
        lower_builtin(op, SeriesType, types.Any)(lower_impl)
        lower_builtin(op, types.Any, SeriesType)(lower_impl)
        whecp__tcddg = create_overload_cmp_operator(op)
        overload(op, no_unliteral=True)(whecp__tcddg)


_install_cmp_ops()


def install_arith_ops():
    for op in (operator.add, operator.sub, operator.mul, operator.truediv,
        operator.floordiv, operator.mod, operator.pow):
        whecp__tcddg = create_overload_arith_op(op)
        overload(op, no_unliteral=True)(whecp__tcddg)


install_arith_ops()
